from hausse.lib.selector import Selector, Pattern
from hausse.lib import selector
from typing import Callable, Iterable, List, Optional, Union

from hausse.lib import Plugin, SelectorPlugin, Element
import logging


class BaseCollection(Plugin):
    """
    BaseCollection
    ==============

    The BaseCollection is the minimal implementation of a Collection.

    It features Element manual addition, Selector-compatible members iteration
    and customizable sorting and ordering.
    """

    def __init__(self, name, sortBy: Union[str, Callable] = None, reverse: bool = False):
        self.name = name
        self.sortBy = sortBy    # TODO: Implement
        self.reverse = reverse  # TODO: Improve implementation

        self._members = []

    def __str__(self) -> str:
        return self.name

    def __len__(self) -> int:
        return len(self._members)

    def __bool__(self) -> bool:
        # Prevent Collection from being considered False event if empty
        # This trick is useful for boolean comparaison in Selectors implementation
        return True

    def __iter__(self):
        # TODO: Improve this with persistent reversed storage
        if self.reverse:
            return iter(reversed(self._members))
        return iter(self._members)

    def _get_selector(self) -> selector.Collection:
        return selector.Collection(self.name)

    def add(self, element: Element):

        if element in self._members:
            logging.debug(f"Element {element._filename} is already in {self.name} Collection.")
            return

        # Adding to internal element list
        self._members.append(element)

        # Create element collections attribute if it does not exists already
        if hasattr(element, "collections"):
            if isinstance(element.collections, list):
                element.collections = set(element.collections)
            elif not isinstance(element.collections, set):
                element.collections = {element.collections}            
        else:
            setattr(element, "collections", set())

        # Reverse-adding collection name and various attributes to element
        element.collections.add(self.name)

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        # Collection setup
        # Creating dedicated attribute in the settings
        if "collections" not in settings:
            settings["collections"] = dict()

        # Collection registration
        settings["collections"][self.name] = self

        # Metadata shortcut
        if self.name in metadata:
            logging.warning(
                f"Attribute {self.name} already exists in metadata, collection shortcut creation skipped."
            )
        else:
            metadata[self.name] = self


class IndexableCollection(BaseCollection):

    def __init__(self, name, indexBy):
        
        super().__init__(name)
        self.indexBy = indexBy
        self._index = dict()

    def __getitem__(self, key) -> Element:
        return self._index[key]

    def get(self, *args, **kwargs) -> Element:
        return self._index.get(*args, **kwargs)

    def add(self, element: Element):

        # Common adding process
        super().add(element)

        # Indexing the element in index dictionnary, if enabled
        if self.indexBy is not None:
            logging.debug(
                f"Collection {self.name} will be indexed using the {self.indexBy} attribute of elements."
            )

            index_name = getattr(element, self.indexBy)

            if index_name in self._index:
                logging.error(
                    f"Element {element._filename} has the same {self.indexBy} attribute as an already registered element, and will not be accessible from the {self.name} collection index."
                )
            elif index_name is None:
                logging.error(
                    f"Element {element._filename} has no {self.indexBy} attribute, and will not be accessible from the {self.name} collection index."
                )
            elif not isinstance(index_name, str):
                logging.error(
                    f"Element {element._filename} has a non-subscritable {self.indexBy} attribute, and will not be accessible from the {self.name} collection index."
                )
            else:
                self._index[index_name] = element


class Collection(IndexableCollection, SelectorPlugin):
    """
    Collections
    ===========

    Groups Elements into an iterable collection.

    This plugin gathers Elements, generally according to their path, and then
    allows to iterate on them.

    The criteria for selecting the Elements is provided by a path pattern
    string in `selection` argument. It may also be any Selector object.

    Example
    -------

    To build a static blog, articles documents may be grouped together.
    This creates a object that can be used to iterate over the articles.
    >>> Collection("articles", "articles/*")

    Once a Collection is created, it is stored `settings['collections'][name]`, where `name` is the first and required parameter.

    The Collection is also accessible directly by its name in global `metadata`, which allows usage of these kind of calls in layouts :

    ```handlebars
    {{#each articles}}
        {{> article_template this}}
    {{/each}}
    ```

    Attributes
    ----------
    name : str
        Name of the Collection. This is the only mandatory parameter at initialization.
    selection : Selector | str | iterable[Element]
        A Selector object to indicate which Elements should be added into the Collection.
        It can also be a filepath pattern string, or an iterable over the wanted Elements.
        If `None`, the pattern `"{name}/*`will be used by default, using the Collection's name.
    sortBy : str | Callable
        #TODO: Implement sortBy
        Not yet implemented.
    indexBy : str
        #TODO: Implement indexBy
        Not yet implemented.
    reverse : bool
        #TODO: Implement reverse
        Not yet implemented.
    metadata : dict
        Key-values fields to be added to the Collection.
    **kwargs : dict
        Any additional keyword argument will be added to the metadata of the Collection.
        Keywords arguments are added after `metadata` argument and overwrite its values if keys are present in both.
    """

    # TODO: Implement pattern, sortBy and reverse stuffs
    def __init__(
        self,
        name: str,
        selector: Union[Selector, str, Iterable[Element]] = None,
        indexBy: Optional[str] = None,
        metadata: dict = None,
        **kwargs,
    ):
        IndexableCollection.__init__(self, name, indexBy)
        SelectorPlugin.__init__(self, selector, Pattern(f"{name}/*"))

        for k, v in (metadata or dict() | kwargs).items():
            setattr(self, k, v)

    
    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        super().__call__(elements, metadata, settings)

        # Collection filling
        for element in self.selector(elements, metadata, settings):
            self.add(element)


class Collections(Plugin):
    """
    Collections
    ===========

    Create multiple Collections, by selection and by metadata.

    This plugin will create automatically Collections based on the `collections` metadata of Elements if present.
    """
    
    def __init__(self, collections: dict = {}):

        self.collections = [
            Collection(name, **collection)
            for name, collection in collections.items()
        ]

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        # Collections setup
        if not "collections" in settings:
            settings["collections"] = dict()

        # Collections registration
        # TODO: Possible optimization by not executing as plugin but filling all by once in elements loop ?
        for collection in self.collections:
            collection(elements, metadata, settings)

        # Parsing elements 
        for element in elements:

            # Check collection attr in element
            if hasattr(element, "collections"):    

                # Fix single intially set collection
                if isinstance(element.collections, list):
                    element.collections = set(element.collections)
                elif not isinstance(element.collections, set):
                    element.collections = {element.collections}

                # Iterating on set element collections
                for name in element.collections:

                    # Auto-creation
                    if name not in settings["collections"]:
                        BaseCollection(name)(elements, metadata, settings)

                    settings["collections"][name].add(element)


from json import dumps
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable
from inspect import isclass

class BaseSelector(ABC):
    @abstractmethod
    def __init__(self):
        """Plugin initialisation"""
        raise NotImplementedError

    @abstractmethod
    def __call__(self, elements: list, metadata: dict, settings: dict):
        """Plugin work"""
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.selection})"

    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__ and self.selection == other.selection

    def save(self) -> dict:
        """Plugin str serialization"""
        return {self.__class__.__name__.lower(): self.selection}


class Pattern(BaseSelector):

    def __init__(self, pattern):
        self.selection = pattern

    def __call__(self, elements: list, metadata: dict, settings: dict):
        return (element for element in elements if element._path.match(self.selection))

class Extensions(BaseSelector):
    
    def __init__(self, *extensions):
        # As we compare filename extensions, we can rely on `.lower()` without considering special unicode combinaisons
        self.selection = list(map(lambda x: x.lstrip(".").lower(), extensions))

    def __call__(self, elements: list, metadata: dict, settings: dict):
        return (
            element
            for element in elements
            if element._path.suffix.lstrip(".").lower() in self.selection
        )

class Elements(BaseSelector):
    def __init__(self, elements):
        self.selection = elements

    def __call__(self, elements: list, metadata: dict, settings: dict):
        return iter(self.selection)

class Collection(BaseSelector):
    def __init__(self, collection):
        self.selection = collection
    
    def __call__(self, elements: list, metadata: dict, settings: dict):
        return iter(settings['collections'][self.selection])

class All(BaseSelector):
    def __init__(self, _ = None):
        self.selection = "all"  # Used only for json save
    
    def __call__(self, elements: list, metadata: dict, settings: dict):
        return iter(elements)


selectors = {k: v for k, v in globals().items() if isclass(v) and v.__module__ == "hausse.lib.selector" and k != "BaseSelector"}

def Selector(selection) -> BaseSelector:

    if isinstance(selection, BaseSelector):
        # Selector(Selector()) == Selector()
        return selection

    if isinstance(selection, dict):
        # TODO: Support selectors unions
        selector = list(selection.keys())[0]
        if selector in selectors:
            return selectors[selector](selection[selector])

    if hasattr(selection, "_get_selector"):
        # This case can accept a Collection object, and will return a selector
        # over the current Collection members
        return selection._get_selector()

    if isinstance(selection, str):
        # If selection is a string, assuming that it is a filepath pattern
        return Pattern(selection)

    if isinstance(selection, Iterable):
        # If selection is a non-string iterable, assuming it is an iterable
        # over already chosen Elements
        return Elements(selection)



import logging
from typing import List
from hausse.lib import Plugin, Element
from hausse.lib.selector import Selector

class Metadata(Plugin):

    """
    Metadata
    ========

    Adds metadata to a selection of elements.

    For each Element in the provided `selection`, each `new_metadata` items
    will be added as object arguments.

    Parameters
    ----------
    - `selection` (Selector) :
        Elements to edit.
    - `new_metadata` (dict) :
        A dictionnary of metadata elements added to the selected collection.
    - `replace` (bool, `False` by default) :
        If True, allows the plugin to overwrite existing metadata in elements.
        
    """

    def __init__(self, selection, new_metadata, replace : bool = False):
        
        self.selection = Selector(selection)
        self.new_metadata = new_metadata
        self.replace = replace

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        selection = self.selection(elements, metadata, settings)

        for element in selection:
            for key, value in self.new_metadata.items():
                if not self.replace and hasattr(element, key):
                    # Prevent metadata overwritting
                    continue
                setattr(element, key, value)


# import pytest
# import logging


# from hausse.plugins import Metadata

# LOG = logging.getLogger(__name__)


# def test_metadata_warn_underscore(caplog):

#     metadata = {"a": 1}
#     new = {"_b": 2}
    
#     m = Metadata(new)

#     with caplog.at_level(logging.WARNING):
#         m(None, metadata, None)

#     # Check metadata values
#     assert metadata.get("a") == 1
#     assert metadata.get("_b") == 2

#     # Check logging
#     assert "_b" in caplog.text

# def test_metadata_default(caplog):

#     metadata = {"a": 1, "bar": 2}
#     new = {"foo": 3, "bar": 4}
    
#     m = Metadata(new)

#     with caplog.at_level(logging.ERROR):
#         m(None, metadata, None)

#     # Check metadata values
#     assert metadata.get("a") == 1
#     assert metadata.get("bar") == 2
#     assert metadata.get("foo") == 3

#     # Check logging
#     assert "bar" in caplog.text


# def test_metadata_replace():

#     metadata = {"a": 1, "bar": 2}
#     new = {"foo": 3, "bar": 4}
    
#     m = Metadata(new, replace=True)
#     m(None, metadata, None)

#     assert metadata.get("a") == 1
#     assert metadata.get("bar") == 4
#     assert metadata.get("foo") == 3


import os, shutil, logging

def clean_dir(dir):
    for filename in os.listdir(dir):
        file_path = os.path.join(dir, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            logging.error(f"Failed to delete {file_path} during clean-up. Reason: {e}.")



from hausse.lib import Element
import pytest
import logging


from hausse.plugins import Markdown

LOG = logging.getLogger(__name__)


def test_markdown():

    elt = Element("a.md")
    elt._contents = "# H1\n\nLorem Ipsum\n"

    M = Markdown()

    M([elt], None, None)

    assert elt._contents == '<h1 id="h1">H1</h1>\n\n<p>Lorem Ipsum</p>\n'


def test_markdown_metadata():

    elt = Element("a.md")
    elt._contents = "---\nfoo: test a\nbar: test b\n---\n# H1\n\nLorem Ipsum\n"

    M = Markdown()

    M([elt], None, None)

    assert elt._contents == '<h1 id="h1">H1</h1>\n\n<p>Lorem Ipsum</p>\n'
    assert getattr(elt, "foo") == 'test a'
    assert getattr(elt, "bar") == 'test b'




import os
from pathlib import Path
import shutil
from typing import List

from hausse.lib import PathPlugin, Element
from hausse.utils import Keys


class Assets(PathPlugin):
    """
        Assets
        ======

        Copy all files from an asset directory, bypassing all others plugins work.

        This plugin is useful for copying static files like CSS stylesheets.
    """

    def __init__(self, path: str = "assets"):
        super().__init__(path)

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):
        for element in os.listdir(self.path):
            s = os.path.join(self.path, element)
            d = os.path.join(settings[Keys.DIST], element)

            if os.path.isdir(s):
                if os.path.exists(d):
                    shutil.rmtree(d)
                shutil.copytree(s, d, False, None)
            else:
                shutil.copy2(s, d) 


from hausse.lib.element import Element
from typing import List
from hausse.lib import SelectorPlugin

class Drop(SelectorPlugin):
    """
    Drop specified elements from the current loaded elements.
    """

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):
        for element in list(self.selector(elements, metadata, settings)):
            elements.remove(element)
    


import logging
import json
import yaml
import os
from importlib import import_module
from typing import Iterable, Union

from .lib import Plugin, Element
from .utils import clean_dir, Keys, Defaults
from pathlib import Path
from copy import deepcopy


class Hausse(object):
    """
    Hausse
    ======

    A pluggable static project generator.

    Hausse can build a static project using a customized list of plugins.

    A Hausse object should be initialized, populated by initialized plugins
    with `use()` method, then executed with the `build()` method. Note that
    without any plugins, Hausse will do no more thant copying files from the
    source directory (`src` by default) to the output directory (`dist` by
    default). All processing is done by the plugins.

    Note also that all Hausse's method returns `self`, for convenient methods
    call chaining.

    Parameters
    ----------
    - `base` ( str ) :
        Working directory or hausse.json file path. Current folder by default.
        If None and hausse.json available in current folder, it will be used.
    - `elements` ( list[ Element ] ) :
        Additional elements to be add to the project
    - `metadata` ( dict ) :
        Additional metadata to be add to the project
    - `settings` ( dict ) :
        Additional settings to be add to the project
    - `**kwargs` ( dict ) :
        Additional settings to be add to the project. Overwrite `settings`.


    Examples
    --------

    >>> Hausse().build()
    # Copy the content of `./src` to `./dist`

    >>> Hausse().use(Markdown()).build()
    # Parses markdown files from `./src` to HTML files in `./dist`.

    >>> Hausse("website").use(Markdown()).build()
    # Same as before, but using `./website/src` and `./website/dist`

    >>> Hausse("website", metadata={"foo": "bar"}).use(Markdown()).build()
    # Same as before, but with "foo" metadata accessible everywhere.


    Attributes
    ----------
    - `_plugins` ( list[ Plugin ] ) :
        Registered plugins that will be called in `build()` method.
    - `elements` ( list[ Element ] ) :
        Elements (files, mainly) loaded from source to be processed.
    - `metadata` ( dict ) :
        Global metadata accessible from any Element.
    - `settings` ( dict ) :
        Technical storage dictionary for plugins interactions.
    """

    def __init__(
        self,
        base_dir: str = None,
        elements: list[Element] = None,
        metadata: dict = None,
        settings: dict = None,
        **kwargs
    ):

        # Loaded plugins list
        self._plugins = list()

        # Data
        self.elements = elements or list()
        self.metadata = metadata or dict()
        self.settings = Defaults.SETTINGS | (settings or dict()) | kwargs

        # Base directory
        base_path = Path(base_dir) if base_dir else Path(Defaults.BASE)
        self.settings[Keys.BASE] = base_path.parent if base_path.is_file() else base_path
            

    def source(self, src: str = Defaults.SRC):
        """Sets the source files directory path. `src` by default."""

        self.settings[Keys.SRC] = Path(src or Defaults.SRC)

        return self

    def destination(self, dist: str = Defaults.DIST):
        """Sets the output directory path. `dist` by default."""

        self.settings[Keys.DIST] = Path(dist or Defaults.DIST)

        return self

    def clean(self, clean: bool = Defaults.CLEAN):
        """Toggle output directory cleaning. False by default."""

        self.settings[Keys.CLEAN] = clean if clean is not None else Defaults.CLEAN

        return self

    def use(self, plugin: Union[Iterable[Plugin], Plugin]):
        """Register a plugin or a list of plugins in the Hausse project plugins list to be used."""

        if isinstance(plugin, list):
            self._plugins += plugin
        else:
            self._plugins.append(plugin)

        return self

    def build(self):
        """Build the project. All the actual processing logic is run in this method scope."""

        # Saving original working directory
        owd = os.getcwd()

        # Set working directory
        os.chdir(self.settings.get(Keys.BASE, Defaults.BASE))

        # Cleaning dist directory
        if self.settings.get(Keys.CLEAN, Defaults.CLEAN):
            clean_dir(self.settings.get(Keys.DIST, Defaults.DIST))

        # Load all source files if source folder is defined
        src = self.settings.get(Keys.SRC)
        if src:
            self.elements += [
                Element(p.relative_to(src), src, self.metadata)
                for p in src.rglob("*")
                if p.is_file()
            ]

        # Apply all plugins work
        for plugin in self._plugins:
            plugin(self.elements, self.metadata, self.settings)

        # Saving built files
        dist = self.settings.get(Keys.DIST)
        if dist:
            for document in self.elements:

                # Arborescence creation
                document_path = Path(os.path.join(dist, document._path))
                document_path.parent.mkdir(parents=True, exist_ok=True)

                # Saving built file
                with open(os.path.join(dist, document._path), "w") as file:
                    file.write(str(document))

        # Restoring original working directory
        os.chdir(owd)

        return self

    def save(self, file = None, mode = None, hidden: bool = False):
        """Save the current configuration to a json file, which can be used in command line"""

        if file is None:
            file = Defaults.FILES[0]

        if hidden and not file.startswith('.'):
            file = '.' + file

        settings = dict()
        if str(self.settings[Keys.SRC]) != Defaults.SRC:
            settings[Keys.SRC] = str(self.settings[Keys.SRC])
        if str(self.settings[Keys.DIST]) != Defaults.DIST:
            settings[Keys.DIST] = str(self.settings[Keys.DIST])
        if self.settings[Keys.CLEAN] != Defaults.CLEAN:
            settings[Keys.CLEAN] = str(self.settings[Keys.CLEAN])
        settings[Keys.PLUGINS] = {plugin.__class__.__name__: plugin.save() for plugin in self._plugins}

        with open(file, 'w', encoding='utf-8') as f:
            if mode == 'json' or not mode and file.endswith('.json'):
                json.dump(settings, f, ensure_ascii=False, indent=4)

            if mode == 'yaml' or mode == 'yml' or not mode and file.endswith('.yml') or not mode and file.endswith('.yaml'):
                yaml.dump(settings, f, allow_unicode=True)


    def load(self, file = None, mode = None):
        """Loads a `hausse.json` settings file"""

        if file is None:
            for default in Defaults.FILES:
                if (self.settings[Keys.BASE] / Path(default)).exists():
                    file = self.settings[Keys.BASE] / Path(default)
                    break

        file = Path(file)

        try:
            with open(file) as settings:
                
                if mode == 'json' or not mode and file.suffix == '.json':
                    settings = json.load(settings)

                elif mode == 'yaml' or mode == 'yml' or not mode and file.suffix in ['.yml', '.yaml']:
                    settings = yaml.load(settings)                

                self.source(settings.get(Keys.SRC))
                self.destination(settings.get(Keys.DIST))
                self.clean(settings.get(Keys.CLEAN))

                plugins = import_module("hausse.plugins")
                
                for name, kwargs in settings.get("plugins", []).items():
                    plugin = getattr(plugins, name)
                    if plugin:
                        self.use(plugin(**kwargs))
        except:
            logging.warn(f"Failed to load {file} settings file.")


    
    # Aliases
    src = source
    dist = destination
    dest = destination



from typing import Callable, List, Union, Optional

from hausse.lib import Plugin, Element
import logging
import csv
import sys


class MultiLanguage(Plugin):
    """
        Multi Language
        ==============

        Support elements in multiple languages.
    """

    KEY = "lang"

    def __init__(self, default_language: str = None, languages = None):
        self.default = default_language
        self.languages = None

    def __iter__(self):
        return iter(self.languages)

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        for element in elements:
            if hasattr(element, self.KEY):
                language = getattr(element, self.KEY)



import logging
from hausse.plugins.layout.handlebars import Handlebars

LAYOUTS = {
    "handlebars": Handlebars
}

# TODO: Merge with LayoutPlugin and make it a metaclass
def Layouts(engine: str = "handlebars", *args, **kwargs):
    plugin = LAYOUTS.get(engine.lower())

    if plugin is None:
        logging.error(f"Engine {engine} not in available layout engines : {', '.join(list(LAYOUTS.keys()))}")
    
    else:
        return plugin(*args, **kwargs)


import logging
from typing import List
from pathlib import Path
from hausse.lib import Plugin, Element

class Index(Plugin):

    """
    Index
    =====

    Adds index in directory

    Useful to create an index page ex-nihilo without specific content.

    Parameters
    ----------
    - `directory` (Path) :
        Directory where to create an index file. `.` by default.
    - `filename` (str) :
        Index filename. `index.html` by default.
    - `metadata` (dict) :
        Metadata passed to the new index Element.
    - `**kwargs` :
        Metadata passed to the new index Element.
    """

    def __init__(self, directory : Path = None, filename : str = "index.html", metadata : dict = None, **kwargs):
        
        self.directory = Path(directory) or Path(".")
        self.filename = filename
        self.metadata = metadata or dict() | kwargs

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        elements.append(Element(self.directory / Path(self.filename), global_metadata=metadata, **self.metadata))
    


import pytest

from pathlib import PurePath

from hausse.plugins import Collections, Collection
from hausse.lib import Element


def test_collection_add():
    

    d1 = Element(PurePath("a/b/c"))

    collection = Collection("test")

    collection.add(d1)

    assert d1 in collection
    assert collection.name in d1.collections


def test_collections_init():

    # TODO: test more accurately pattern matching

    c = Collections({
        "test_a": {
            "pattern": "a/**/*"
        },
        "test_a/b": {
            "pattern": "a/b/*",
            "reverse": True
        }
    })

    assert len(c.collections) == 2


def test_collections_call():

    # TODO: Test Collections sub-methods instead of main call method
    # TODO: Test more accurately collection members indexing and sorting

    elements = []
    settings = dict()
    metadata = dict()

    d1 = Element(PurePath("a/b/c"), global_metadata=metadata)
    d2 = Element(PurePath("a/b/f"), global_metadata=metadata)
    d3 = Element(PurePath("a/k/c"), global_metadata=metadata)
    d4 = Element(PurePath("a/b/d/c"), global_metadata=metadata)

    elements.append(d1)
    elements.append(d2)
    elements.append(d3)
    elements.append(d4)

    c = Collections({
        "test_a": {
            "selector": "**/c"
        },
        "test_b": {
            "selector": "a/b/*",
            "reverse": True
        }
    })

    c(elements, metadata, settings)

    # Check plugin settings entry
    assert "collections" in settings

    c1 = settings.get("collections", {}).get("test_a")
    c2 = settings.get("collections", {}).get("test_b")

    # Check collection creation
    assert isinstance(c1, Collection)
    assert isinstance(c2, Collection)

    # Check collection setup
    assert getattr(c1, "name") == "test_a"
    assert getattr(c2, "name") == "test_b"

    # Check correct collections assignement
    assert str(c1) in getattr(d1, "collections", [])
    assert str(c1) in getattr(d3, "collections", [])
    assert str(c1) in getattr(d4, "collections", [])
    assert str(c2) in getattr(d1, "collections", [])
    assert str(c2) in getattr(d2, "collections", [])
    assert str(c2) not in getattr(d3, "collections", [])
    assert str(c2) not in getattr(d4, "collections", [])

    # Check global metadata binding
    assert metadata.get("test_a") == c1
    assert metadata.get("test_b") == c2

    # Check global metadata access
    assert getattr(d4, "test_b") == c2

    # Check collection elements order
    assert list(c1) == [d1, d3, d4]
    assert list(c2) == [d2, d1]



from typing import List
from markdown2 import markdown

from hausse.lib import SelectorPlugin, Element
from hausse.lib.selector import Extensions


class Markdown(SelectorPlugin):
    """
    Markdown
    ========

    Parse Markdown files to HTML.

    Parameters
    ----------
    - `selection` :
        Files to parse. By default, all files with `.md` or `.markdown` extensions are selected.
    - `extras` (List[str]) :
        Extras settings passed to the markdown processor.
    """

    default_extras = ['cuddled-lists', 'fenced-code-blocks', 'footnotes', 'header-ids', 'markdown-in-html', 'metadata', 'noreferrer', 'tag-friendly', 'task_list']

    def __init__(self, selection = None, extras: List[str] = None):

        super().__init__(selection, Extensions("md", "markdown"))
        self.extras = extras


    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        for element in self.selector(elements, metadata, settings):
            m = markdown(element._contents, extras=self.extras or self.default_extras)
            element._contents = str(m)
            element._update_metadata(m.metadata)


from typing import List
from hausse.lib import SelectorPlugin, Element


class Keep(SelectorPlugin):
    """
    Keep only specified elements and drop everything else.
    """

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):
        elements[:] = self.selector(elements, metadata, settings)
    


import pytest

from pathlib import PurePath

from hausse.plugins import Keep
from hausse.lib import Element


def test_keep_simple_pattern():

    d1 = Element(PurePath("a/b/c"))
    d2 = Element(PurePath("a/b/f"))
    d3 = Element(PurePath("a/k/c"))
    d4 = Element(PurePath("a/k/d/c"))

    elements = [d1, d2, d3, d4]

    K = Keep("a/b/c")

    K(elements, None, None)

    assert len(elements) == 1
    assert d1 in elements
    assert d2 not in elements
    assert d3 not in elements
    assert d4 not in elements


def test_keep_wildcard_pattern():

    d1 = Element(PurePath("a/b/c"))
    d2 = Element(PurePath("a/b/f"))
    d3 = Element(PurePath("a/k/c"))
    d4 = Element(PurePath("a/k/d/c"))

    elements = [d1, d2, d3, d4]

    K = Keep("a/*/c")

    K(elements, None, None)

    assert len(elements) == 2
    assert d1 in elements
    assert d2 not in elements
    assert d3 in elements
    assert d4 not in elements


def test_keep_multiple_wildcard_pattern():

    d1 = Element(PurePath("a/b/c"))
    d2 = Element(PurePath("a/b/f"))
    d3 = Element(PurePath("a/k/c"))
    d4 = Element(PurePath("a/k/d/c"))

    elements = [d1, d2, d3, d4]

    K = Keep("**/c")

    K(elements, None, None)

    assert len(elements) == 3
    assert d1 in elements
    assert d2 not in elements
    assert d3 in elements
    assert d4 in elements



import os
from hausse.lib.plugin import Plugin
import pytest
from pathlib import Path
from hausse import Hausse

def test_init_defaults():

    h = Hausse()

    assert isinstance(h._plugins, list)
    assert isinstance(h.elements, list)
    assert isinstance(h.settings, dict)
    assert isinstance(h.metadata, dict)
    assert Path(h.settings.get("base_dir")) == Path(".")
    assert Path(h.settings.get("src_dir")) == Path("src")
    assert Path(h.settings.get("dist_dir")) == Path("dist")


def test_init_additional_settings():

    h = Hausse(arg1="Foo", arg2="Bar")
    
    assert h.settings.get("arg1") == "Foo"
    assert h.settings.get("arg2") == "Bar"


def test_init_paths():

    h = Hausse(base_dir=Path("a/b/c"), src_dir="d/e/f", dist_dir=Path("g/h"))
    
    assert Path(h.settings.get("base_dir")) == Path("a/b/c")
    assert Path(h.settings.get("src_dir")) == Path("d/e/f")
    assert Path(h.settings.get("dist_dir")) == Path("g/h")


def test_source():

    h = Hausse()

    h.source("d/e/f")

    assert Path(h.settings.get("src_dir")) == Path("d/e/f")


def test_dist():

    h = Hausse()

    h.dist("g/h")

    assert Path(h.settings.get("dist_dir")) == Path("g/h")


def test_clean():

    h = Hausse()

    h.clean()

    # If `clean()` is called without argument, then Hausse should clean dist
    assert h.settings.get("clean") is True

    h.clean(False)

    # But it is still possible to explicitly disable cleaning
    assert h.settings.get("clean") is False


def test_use():

    class FooPlugin(Plugin):
        def __init__(self):
            pass
        def __call__(self):
            pass

    h = Hausse()
    p1 = FooPlugin()
    p2 = FooPlugin()

    h.use(p1)
    h.use(p2)

    assert h._plugins[0] is p1
    assert h._plugins[1] is p2


def test_build():

    H = Hausse()

    # Using None to disable Elements autoload and write
    H.source(None)
    H.dist(None)

    # Defining minimalistic dummy plugins
    p1 = lambda e, m, s: s.update({"test": "good"})
    p2 = lambda e, m, s: s.update({"test": s["test"]+"!"})

    H.use(p1).use(p2).use(p2)

    # Saving original working directory
    owd = os.getcwd()

    H.build()
    
    # Check non modified current directory
    assert os.getcwd() == owd

    # Check if internal variables have been modified accurately
    assert H.settings["test"] == "good!!"



import logging
from pathlib import Path, PurePath
from typing import Optional

class Element(object):
    """
        The Element stores processed files content and dedicated metadata.

        Parameters
        ----------
        path : PurePath
            Relative path of file location
        source : Path
            Path of the source folder from where the file is loaded
        global_metadata : dict
            Reference to the Hausse global metadata dict, for convenient access
        **kwargs :
            Any additional metadata to be added to the Element attributes
    """

    def __init__(self, path: PurePath, source: Optional[Path] = None, global_metadata: Optional[dict] = None, **kwargs):

        self._path = Path(path)
        self._filename = self._path.name
        self._global_metadata = global_metadata
        
        # Metadata
        for k, v in kwargs.items():
            setattr(self, k, v)

        # Content
        if source is not None and self._path is not None:
            self._load(source)


    def __getattr__(self, key):
        try:
            return self._global_metadata[key]
        except KeyError:
            raise AttributeError(f"'{self._filename}' Element has no attribute '{key}', nor has the global metadata")
        except TypeError:
            raise AttributeError(f"'{self._filename}' Element has no attribute '{key}', and global metadata is not accessible")
        

    def __str__(self):
        return self._contents

    
    def _metadata(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}


    def _update_metadata(self, new: dict):
        for key, value in new.items():
            if key.startswith('_'):
                logging.warning(f"Adding a hidden metadata {key} in {self._filename} element.")
            if hasattr(self, key):
                logging.warning(f"Overwritting existant {key} metadata in {self._filename} element.")
            setattr(self, key, value)

    def _load(self, source):
        with open(Path(source) / Path(self._path), "r") as file:
            self._contents = file.read()


import glob
from pathlib import Path, PurePath
import os
from typing import List

from hausse.lib import Element, PathPlugin

class DiscoverPartials(PathPlugin):
    """
    Register layouts partials before layout engine plugin execution.
    """

    def __init__(self, path: str = "partials", pattern: str = "*.hbs"):
        super().__init__(path)
        self.pattern = pattern

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):
        
        settings['partials'] = dict()

        for f in self.path.rglob(self.pattern):

            with open(f, "r") as p:

                settings['partials'][PurePath(f).stem] = p.read()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Optional, Union
from inspect import signature, Parameter

from hausse.utils import Defaults
from .selector import Selector, All


class Plugin(ABC):
    """
    Hausse plugin structure
    """

    @abstractmethod
    def __init__(self):
        """Plugin initialisation"""
        raise NotImplementedError

    @abstractmethod
    def __call__(self, elements: list, metadata: dict, settings: dict):
        """Plugin work"""
        raise NotImplementedError

    def save(self):
        """Returns a JSON-like dict representing the plugin instance config. Used by `Hausse.save()`."""
        defaults = {k: v.default for k, v in signature(self.__init__).parameters.items() if v.default is not Parameter.empty}
        # Return all key-values from the plugin except
        # - technic internal keys (identified by leading underscore)
        # - default values (identified by plugin init signature inspection)
        # - and empty values corresponding to None defaults (empty lists initialized from None arguments)
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_") and v != defaults.get(k) and (v or defaults.get(k))}


class SelectorPlugin(Plugin):
    """
    Plugin template with a Selector parameter.

    Selector can be a file path pattern, an iterator over an Elements list, or any other Selector object.
    """

    def __init__(self, selector: Union[str, Iterable, Selector], selector_by_default = None):
        self.selector = Selector(selector or selector_by_default)
        self._selector_by_default = selector_by_default

    def save(self):
        d = super().save()
        if self.selector == self._selector_by_default:
            del d['selector']
        else:
            d['selector'] = d['selector'].save()
        return d
        
class PathPlugin(Plugin):
    """
    Plugin template with a Path parameter.

    PathPlugin should be used when manipulating files outside of source folder.
    """

    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def save(self) -> dict:
        d = super().save()
        d['path'] = str(d['path'])
        if d['path'] == signature(self.__init__).parameters['path'].default:
            del d['path']
        return d


class LayoutPlugin(PathPlugin, SelectorPlugin):
    """
    Hausse layout plugin structure
    """

    def __init__(self, path: str = Defaults.LAYOUTS, default: str = None, selector: Union[str, Iterable, Selector] = None, options: dict = None, **kwargs):
        """
        # TODO: Update documentation
        Parameters
        ----------
        directory: str
            Layouts folder path
        default: str
            Default layout file that should be used if none is specified
        pattern: str
            Matching pattern for the elements to be processed
        options: dict
            Options passed to the layout engine
        **kwargs: dict
            Additional options passed to the layout engine
        """
        SelectorPlugin.__init__(self, selector, All())
        PathPlugin.__init__(self, path)
        self.default = default
        self.options = options or dict() | kwargs


from typing import Callable, List, Union, Optional

from hausse.lib import Plugin, Element
import logging
import csv
import sys


class CSVExpand(Plugin):
    """
        CSVExpand
        =========

        Loads CSV files as collections of elements.
    """

    def __init__(self, pattern = None, delimiter = ",", contents_column = None, filename_column = None):
        
        self.pattern = pattern
        self.delimiter = delimiter
        self.contents_column = contents_column
        self.filename_column = filename_column


    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        csv.field_size_limit(sys.maxsize)

        # Parsing elements
        for element in elements:

            # Valid selected CSV file
            if self.pattern and element._path.match(self.pattern) or self.pattern is None and element._path.suffix.lower() == ".csv":

                # Parsing CSV file
                records = csv.reader(element._contents.splitlines(), delimiter=self.delimiter)

                # Header line
                headers = next(records)

                # Iterating records
                for i, row in enumerate(records):

                    # Parsing CSV data into metadata
                    metadata = dict(zip(headers, row))

                    # Extracting main column contents
                    contents = metadata.pop(self.contents_column) if self.contents_column else None

                    # New filename building
                    filename = metadata.get(self.filename_column) if self.filename_column else str(i)
                    filepath = element._path.with_name(filename)
                    
                    elements.append(Element(filepath, contents, **metadata))

                # Removing original CSV file
                del element


from pathlib import Path


# Main keys

class Keys:
    BASE = "base_dir"
    CLEAN = "clean"
    DIST = "dist_dir"
    SRC = "src_dir"
    PLUGINS = "plugins"


# Default values

class Defaults:
    FILES = ["hausse.json", "hausse.yaml", "hausse.yml", ".hausse.json", ".hausse.yaml", ".hausse.yml"]
    BASE = "."
    CLEAN = True
    DIST = "dist"
    SRC = "src"
    LAYOUTS = "layouts"
    SETTINGS = {Keys.BASE: Path(BASE), Keys.SRC: Path(SRC), Keys.DIST: Path(DIST), Keys.CLEAN: False}




import logging
from typing import List
from hausse.lib import SelectorPlugin, Element
from hausse.lib.selector import Selector

class Relations(SelectorPlugin):
    """
    Relations
    =========

    Create relations between elements based on a foreign key.

    For each Element in `selector`, this plugin will replace its `key` attribute, if present, by references to the corresponding Element in provided `collection`.

    # TODO: Finish documentation
    """

    def __init__(self, selector, collection, key = None):
        
        super().__init__(selector)
        self.collection = collection if isinstance(collection, str) else collection.name
        self.key = key


    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        if self.collection in settings["collections"]:

            collection = settings["collections"][self.collection]
            
            for element in self.selector(elements, metadata, settings):

                key = getattr(element, self.key or self.collection, None)

                if key:

                    if isinstance(key, list):
                        setattr(element, self.key or self.collection, [collection.get(elt, elt) for elt in key])
                        
                    else:
                        setattr(element, self.key or self.collection, collection.get(key, key))
        
        else:

            if "collections" not in settings:
                logging.error(f"Relations plugin must be used after Collections plugin.")

            logging.error(f"Relations did not found {self.collection} collection in project metadata.")


"""
Plugins
=======

This module contains all available Hausse plugins.

"""

from .assets import Assets

from .csv_expand import CSVExpand

# Elements organization plugins
from .elements.drop import Drop
from .elements.keep import Keep
from .elements.collection import Collection, Collections
from .elements.relations import Relations
from .elements.index import Index

# Layouts-related plugins
from .layout.handlebars import Handlebars
from .layout.layouts import Layouts
from .layout.discover_partials import DiscoverPartials

# Data processing plugins
from .data.markdown import Markdown

# Metadata processing plugins
from .metadata.metadata_markdown import MetadataMarkdown


import argparse
from pathlib import Path

from .utils import Defaults, Keys
from .hausse import Hausse

parser = argparse.ArgumentParser(prog="hausse", description="Modulable static project builder")

parser.add_argument("Path", type=str, nargs="?", help="Project folder or hausse.json file path")
parser.add_argument("-b", "--build", action="store_true", help="Builds the project")

parser.parse_args()

path = Path(parser.Path or ".")

project = Hausse(path)

if path.is_file and path.exists():
    project.load(path)
else:
    project.load()

if parser.build:
    project.build()




from typing import List
import markdown2

from hausse.lib import SelectorPlugin, Element
from hausse.lib.selector import All


class MetadataMarkdown(SelectorPlugin):
    """
    MetadataMarkdown
    ================

    Parse elements metadata as Markdown to HTML.

    This plugin accepts any number of keys at initialization. Then, it will
    iterate over Elements and search in the metadata of each of them the
    values corresponding to the provided keys. Finally, it parse theses
    values as markdown documents to html documents.

    Values are modified in place. No Element is created by this plugin.

    An optional keyword argument, `selection`, can be used to limit this
    plugins cope to a subset of Elements. By default, all Elements of the
    project will be processed.

    Parameters
    ----------
    - `*args` ( str ) :
        Metadata keys containing the string to be parsed as metadata
    - `selection` ( Selector | str | Iterable[ Element ] ) :
        If provided, indicates which Elements should be processed.
        By default, all Elements are processed.
        Note that this parameter must be passed explicitly.

    Example
    -------
    ```python
        >>> element = Element("test")
        >>> element.foo = "# Foo Example\\nThis is a **nice** example."
        >>> plugin = MetadataMarkdown("foo")
        >>> plugin([element], dict(), dict())
        >>> print(element.foo)
        <h1 id="foo-example">Foo Example</h1>
        <p>This is a <b>nice</b> example
    ```

    Attributes
    ----------
    - `keys` ( List[ str ] ) :
        Metadata keys containing the string to be parsed as metadata
    - `selection` ( Selector ) :
        If provided, indicates which Elements should be processed.
        By default, all Elements are processed.
        Note that this parameter must be passed explicitly.

    """

    extras = ['cuddled-lists', 'fenced-code-blocks', 'footnotes', 'header-ids', 'markdown-in-html', 'noreferrer', 'tag-friendly', 'task_list']

    def __init__(self, selector=None, *args):      
        super().__init__(selector, All())
        self.keys = set(args)


    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        for element in self.selector(elements, metadata, settings):
            
            for key in self.keys:
                
                value = element._metadata().get(key)
                
                if value:
                    setattr(element, key, markdown2.markdown(value, extras=self.extras))

    
    def save(self) -> dict:
        d = super().save()
        if 'keys' in d:
            d['keys'] = list(d['keys'])
        return d


import glob
import logging
import os
import pathlib
from pathlib import Path, PurePath
from typing import List

import pybars

from hausse.lib import LayoutPlugin, Element

class Handlebars(LayoutPlugin):
    """
    Apply handlebars layouts to elements
    """

    def __call__(self, elements: List[Element], metadata: dict, settings: dict):

        # Compiler init        
        compiler = pybars.Compiler()

        # Partials compilation
        partials = settings.get("partials", dict())
        for name, partial in partials.items():
            partials[name] = compiler.compile(partial)

        # Templates compilation
        templates = settings.get("templates", dict())
        for file in self.path.rglob("*.hbs"):
            with open(file, "r") as layout:
                if PurePath(file).name in templates:
                    logging.warning(f"A layout file named {PurePath(file).name} has already been registered. The new one is skipped.")
                else:
                    templates[PurePath(file).name] = compiler.compile(layout.read())

        for element in self.selector(elements, metadata, settings):

            # Template selection
            template = templates.get(getattr(element, 'layout', self.default))

            if template is None:
                logging.info(f"Element {element._filename} with no defined layout is skipped.")

            else:
                # Render
                element._contents = template(element, partials=partials)
                element._path = element._path.with_suffix('.html')
